==> vitunes-1.0.1.tgz: vitunes-1.0.1/paint.c <==
/*
 * Copyright (c) 2010 Ryan Flannery <ryan.flannery@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "paint.h"

/* global colors object */
_colors colors;

/* This is used to get which field in the playing file to display on a given
 * paint of the player window.
 * The field displayed rotates between artist, album, and title, changing
 * every 3 seconds.
 */
char *
player_get_field2show(const meta_info *mi)
{
   static time_t last_updated = 0;
   static int    index        = 0;
   static int    offset;
   static int    fields[] = { MI_CINFO_ARTIST, MI_CINFO_ALBUM,
                              MI_CINFO_TITLE };

   /* determine which cinfo item to show */
   if (time(NULL) - last_updated >= 3) {
      last_updated = time(NULL);
      index = (index + 1) % 3;
   }

   if (mi->cinfo[ fields[index] ] != NULL)
      return mi->cinfo[ fields[index] ];
   else {
      /* draw the filename if field info not there, but trim it down */
      offset = 0;
      if (strlen(mi->filename) > 49)
         offset += strlen(mi->filename) - 49;

      return mi->filename + offset;
   }
}

/* given a number N, this builds and returns a printf(3) "%Ns" string */
char *
num2fmt(int n)
{
   static char format[255];

   if (n <= 0) {
      endwin();
      errx(1, "num2sfmt: invalid number %d provided", n);
   }

   snprintf(format, sizeof(format), "%%-%d.%ds", n, n);
   return format;
}

/* paint the status bar */
void
paint_status_bar()
{
   swindow    *w;
   static char scratchpad[500];
   char       *focusName;
   int         percent;

   /* NOTE: to right-align the text we snprintf to the scratchpad above and
    * then use the cmd_row_fmt element of ui to right-align the string
    * (scratchpad) in the display.
    * this should be changed... just determine width of display...all fields
    * printed are fixed-width.
    * XXX
    */

   /* determine focus'd window name */
   if (ui.active == ui.library) {
      focusName = "library";
      w = ui.library;
   } else {
      focusName = "playlist";
      w = ui.playlist;
   }

   /* determine how far we've scrolled through the window */
   if (w->nrows == 0)
      percent = 100;
   else
      percent = 100 * (w->voffset + w->crow + 1) / w->nrows;

   /* build the string to print */
   snprintf(scratchpad, sizeof(scratchpad),
      "[%s%s%s] %6d,%-3d %3d%%",
      focusName,
      (ui.active == ui.library ? "" : ":"),
      (ui.active == ui.library ? "" : viewing_playlist->name),
      w->voffset + w->crow + 1,
      w->hoffset,
      percent);

   /* do the printing */
   werase(ui.command);
   wattron(ui.command, COLOR_PAIR(colors.status));
   mvwprintw(ui.command, 0, 0, ui.cmd_row_fmt, scratchpad);
   wattroff(ui.command, COLOR_PAIR(colors.status));
   wrefresh(ui.command);
}

/* paint the player */
void
paint_player()
{
   static char *playmode;
   static char *finfo;
   static int   in_hour;
   static int   in_minute;
   static int   in_second;
   static int   percent, whole;
   int w, h;

   getmaxyx(stdscr, h, w);

   /* if nothing's playing, a shameless plug */
   if (!mp_status.is_playing) {
      werase(ui.player);
      wattron(ui.player, COLOR_PAIR(colors.player));
      mvwprintw(ui.player, 0, 0, num2fmt(w), "vitunes...");
      wattroff(ui.player, COLOR_PAIR(colors.player));
      wrefresh(ui.player);
      return;
   }

   /* determine time into current selection */
   in_hour   = (int)  roundf(mp_status.position / 3600);
   in_minute = ((int) roundf(mp_status.position)) % 3600 / 60;
   in_second = ((int) roundf(mp_status.position)) % 60;

   /* determine percent time into current selection */
   percent = -1;
   if (playing_playlist->files[player.current]->plength_ms > 0)
   {  whole = playing_playlist->files[player.current]->plength_ms / 1000000;
      percent = roundf(100.0 * mp_status.position / whole);
   }

   /* get character for playmode */
   switch (player.mode)
   {  case MP_MODE_LINEAR:
         playmode = "-";
         break;
      case MP_MODE_LOOP:
         playmode = "O";
         break;
      case MP_MODE_RANDOM:
         playmode = "?";
         break;
   }

   /* determine info about song to show */
   finfo = player_get_field2show(playing_playlist->files[player.current]);

   /* draw */
   werase(ui.player);
   wattron(ui.player, COLOR_PAIR(colors.player));
   mvwprintw(ui.player, 0, 0, num2fmt(w), " "); /* this fill the bg color */
   mvwprintw(ui.player, 0, 0,
      "[%s] %8.8s +%2.2d:%2.2d:%2.2d (%d%%) %-49.49s",
      playmode,
      (player.paused ? "-PAUSED-" : ""),
      in_hour, in_minute, in_second,
      percent,
      finfo);
   wattroff(ui.player, COLOR_PAIR(colors.player));
   wrefresh(ui.player);
}

/* paint the library window */
void
paint_library()
{
   char *str;
   int   row, hoff, index;

   /* if library window is hidden, nothing to do */
   if (ui.library->cwin == NULL) return;

   werase(ui.library->cwin);
   wattron(ui.library->cwin, COLOR_PAIR(colors.library));

   for (row = 0; row < ui.library->h; row++)
   {
      index = ui.library->voffset + row;

      /* apply attributes */
      if (index < mdb.nplaylists && mdb.playlists[index] == playing_playlist)
         wattron(ui.library->cwin, COLOR_PAIR(colors.playing));

      if (index < mdb.nplaylists && mdb.playlists[index]->needs_saving)
         wattron(ui.library->cwin, A_BOLD);

      if (row == ui.library->crow)
         wattron(ui.library->cwin, A_REVERSE);

      if (index >= mdb.nplaylists)
         wattron(ui.library->cwin, COLOR_PAIR(colors.tildas));

      /* draw the row */
      if (index >= mdb.nplaylists)
         mvwprintw(ui.library->cwin, row, 0, "~");
      else {
         /* determine horizontal offset */
         str = mdb.playlists[index]->name;
         hoff = ui.library->hoffset;
         if (hoff >= strlen(str))
            hoff = strlen(str);

         /* draw it */
         mvwprintw(ui.library->cwin, row, 0,
            num2fmt(ui.library->w),
            str + hoff);
      }

      /* un-apply attributes */
      if (index >= mdb.nplaylists) {
         wattroff(ui.library->cwin, COLOR_PAIR(colors.tildas));
         wattron(ui.library->cwin, COLOR_PAIR(colors.library));
      }

      if (row == ui.library->crow) {
         wattroff(ui.library->cwin, A_REVERSE);
         wattron(ui.library->cwin, COLOR_PAIR(colors.library));
      }

      if (index < mdb.nplaylists && mdb.playlists[index]->needs_saving) {
         wattroff(ui.library->cwin, A_BOLD);
         wattron(ui.library->cwin, COLOR_PAIR(colors.library));
      }

      if (index < mdb.nplaylists && mdb.playlists[index] == playing_playlist) {
         wattroff(ui.library->cwin, COLOR_PAIR(colors.playing));
         wattron(ui.library->cwin, COLOR_PAIR(colors.library));
      }
   }

   wattroff(ui.library->cwin, COLOR_PAIR(colors.library));
   wrefresh(ui.library->cwin);
}

/* paint the playlist window */
void
paint_playlist()
{
   playlist   *plist;
   bool        hasinfo;
   char       *str;
   int         findex, row, col, colwidth;
   int         xoff, hoff, strhoff;
   int         cattr;


   plist = viewing_playlist;

   werase(ui.playlist->cwin);

   for (row = 0; row < ui.playlist->h; row++) {

      /* get index of file to show */
      findex = row + ui.playlist->voffset;

      /* apply row attributes */
       wattron(ui.playlist->cwin, COLOR_PAIR(colors.playlist));

      if (plist == playing_playlist && findex == player.current)
         wattron(ui.playlist->cwin, COLOR_PAIR(colors.playing));

      if (row == ui.playlist->crow)
         wattron(ui.playlist->cwin, A_REVERSE);

      if (findex >= plist->nfiles)
         wattron(ui.playlist->cwin, COLOR_PAIR(colors.tildas));

      /* draw the row */
      if (findex >= plist->nfiles)
         mvwprintw(ui.playlist->cwin, row, 0, "~");
      else {
         /* this acheives the A_REVERSE attribute spanning the entire row */
         mvwprintw(ui.playlist->cwin, row, 0, num2fmt(ui.playlist->w), " ");

         /* does the file have any meta-info? */
         hasinfo = false;
         for (col = 0; col < _mi_display.nfields; col++) {
            if (plist->files[findex]->cinfo[_mi_display.order[col]] != NULL)
               hasinfo = true;
         }

         /* if there's no meta info, just show filename */
         if (!hasinfo) {
            mvwprintw(ui.playlist->cwin, row, 0, num2fmt(ui.playlist->w),
               plist->files[findex]->filename);
         } else {

            /* loop through all fields of file and display each ... */
            xoff = 0;
            hoff = ui.playlist->hoffset;
            for (col = 0; col < _mi_display.nfields; col++) {

               /* is horizontal offset big enough to skip this field? */
               if (hoff >= _mi_display.widths[col]) {
                  hoff -= _mi_display.widths[col];
                  continue;
               }

               /* field shown off the screen? */
               if (xoff >= ui.playlist->w)
                  continue;

               /* get string to show (str) and horizontal offset to apply (shoff) */
               str = plist->files[findex]->cinfo[_mi_display.order[col]];
               if (str != NULL) {
                  if (hoff > strlen(str))
                     strhoff = strlen(str);
                  else
                     strhoff = hoff;
               }

               /* apply column attribute (only if file is NOT playing) */
               cattr = COLOR_PAIR(colors.cinfos[_mi_display.order[col]]);
               if ((plist != playing_playlist || findex != player.current)
               && colors.cinfos_set[_mi_display.order[col]])
                  wattron(ui.playlist->cwin, cattr);

               /* determine width of this field */
               colwidth = _mi_display.widths[col] - hoff;
               if (xoff + colwidth > ui.playlist->w)
                  colwidth = ui.playlist->w - xoff;

               /* print the column */
               mvwprintw(ui.playlist->cwin, row, xoff, num2fmt(colwidth),
                  (str == NULL ? " " : str + strhoff));

               /* un-apply column attribute */
               if ((plist != playing_playlist || findex != player.current)
               && colors.cinfos_set[_mi_display.order[col]]) {
                  wattroff(ui.playlist->cwin, cattr);
                  wattron(ui.playlist->cwin, COLOR_PAIR(colors.playlist));
               }

               xoff += 1 + colwidth; /* +1 for space between columns */
               hoff = 0;
            }
         }
      }

      /* un-apply row attributes */
      if (findex >= plist->nfiles)
         wattroff(ui.playlist->cwin, COLOR_PAIR(colors.tildas));

      if (row == ui.playlist->crow)
         wattroff(ui.playlist->cwin, A_REVERSE);

      if (plist == playing_playlist && findex == player.current)
         wattroff(ui.playlist->cwin, COLOR_PAIR(colors.playing));

      wattroff(ui.playlist->cwin, COLOR_PAIR(colors.playlist));
   }

   wrefresh(ui.playlist->cwin);
}

/* paint all windows */
void
paint_all()
{
   wattron(stdscr, COLOR_PAIR(colors.bars));
   ui_draw_borders();
   wattroff(stdscr, COLOR_PAIR(colors.bars));

   paint_player();
   paint_status_bar();
   paint_library();
   paint_playlist();
}

/* Paints an error message to the command/status window.  The usage is
 * identical to that of printf(3) (vwprintw(3) actually).
 */
void
paint_error(char *fmt, ...)
{
   va_list ap;

   werase(ui.command);
   wmove(ui.command, 0, 0);
   wattron(ui.command, COLOR_PAIR(colors.errors));

   va_start(ap, fmt);
   vwprintw(ui.command, fmt, ap);
   va_end(ap);

   beep();
   wattroff(ui.command, COLOR_PAIR(colors.errors));
   wrefresh(ui.command);
}

/* Paints an informational message to the command/status window.  The usage
 * is identical to that of printf(3) (vwprintw(3) actually).
 */
void
paint_message(char *fmt, ...)
{
   va_list ap;

   werase(ui.command);
   wmove(ui.command, 0, 0);
   wattron(ui.command, COLOR_PAIR(colors.messages));

   va_start(ap, fmt);
   vwprintw(ui.command, fmt, ap);
   va_end(ap);

   wattroff(ui.command, COLOR_PAIR(colors.messages));
   wrefresh(ui.command);
}

/* Each of these members of the global color object will be
 * run through init_pair(3)
 */
void
paint_setup_colors()
{
   int i;

   /* setup the indices to be used for init_pair/COLOR_PAIR */
   colors.bars = 1;
   colors.errors = 2;
   colors.messages = 3;
   colors.player = 4;
   colors.status = 5;
   colors.library = 6;
   colors.playlist = 7;
   colors.playing = 8;
   colors.tildas = 9;

   /* setup default colors */
   init_pair(colors.bars,     COLOR_WHITE, COLOR_BLACK);
   init_pair(colors.errors,   COLOR_WHITE, COLOR_RED);
   init_pair(colors.messages, COLOR_RED,   COLOR_BLACK);
   init_pair(colors.player,   COLOR_GREEN, COLOR_BLACK);
   init_pair(colors.status,   COLOR_WHITE, COLOR_BLACK);
   init_pair(colors.library,  COLOR_WHITE, COLOR_BLACK);
   init_pair(colors.playlist, COLOR_WHITE, COLOR_BLACK);
   init_pair(colors.playing,  COLOR_GREEN, COLOR_BLACK);
   init_pair(colors.tildas,   COLOR_BLUE,  COLOR_BLACK);

   /* colors for cinfo fields (columns in playlist window) */
   for (i = 0; i < MI_NUM_CINFO; i++) {
      colors.cinfos[i] = 10 + i;
      colors.cinfos_set[i] = false;
   }
}

int
paint_str2item(const char *str)
{
   int i;

   if (strcasecmp(str, "bars") == 0)
      return colors.bars;
   else if (strcasecmp(str, "errors") == 0)
      return colors.errors;
   else if (strcasecmp(str, "messages") == 0)
      return colors.messages;
   else if (strcasecmp(str, "player") == 0)
      return colors.player;
   else if (strcasecmp(str, "status") == 0)
      return colors.status;
   else if (strcasecmp(str, "library") == 0)
      return colors.library;
   else if (strcasecmp(str, "playlist") == 0)
      return colors.playlist;
   else if (strcasecmp(str, "playing") == 0)
      return colors.playing;
   else if (strcasecmp(str, "tildas") == 0)
      return colors.tildas;

   /* if reached here, check cinfo's array */
   for (i = 0; i < MI_NUM_CINFO; i++) {
      if (strcasecmp(str, MI_CINFO_NAMES[i]) == 0)
         return colors.cinfos[i];
   }

   return -1;
}

int
paint_str2color(const char *str)
{
   if (strcasecmp(str, "black") == 0)
      return COLOR_BLACK;
   else if (strcasecmp(str, "red") == 0)
      return COLOR_RED;
   else if (strcasecmp(str, "green") == 0)
      return COLOR_GREEN;
   else if (strcasecmp(str, "yellow") == 0)
      return COLOR_YELLOW;
   else if (strcasecmp(str, "blue") == 0)
      return COLOR_BLUE;
   else if (strcasecmp(str, "magenta") == 0)
      return COLOR_MAGENTA;
   else if (strcasecmp(str, "cyan") == 0)
      return COLOR_CYAN;
   else if (strcasecmp(str, "white") == 0)
      return COLOR_WHITE;
   else
      return -1;
}

==> vitunes-1.0.2.tgz: vitunes-1.0.2/paint.c <==
/*
 * Copyright (c) 2010 Ryan Flannery <ryan.flannery@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "paint.h"

/* global colors object */
_colors colors;

/* This is used to get which field in the playing file to display on a given
 * paint of the player window.
 * The field displayed rotates between artist, album, and title, changing
 * every 3 seconds.
 */
char *
player_get_field2show(const meta_info *mi)
{
   static time_t last_updated = 0;
   static int    index        = 0;
   static int    offset;
   static int    fields[] = { MI_CINFO_ARTIST, MI_CINFO_ALBUM,
                              MI_CINFO_TITLE };

   /* determine which cinfo item to show */
   if (time(NULL) - last_updated >= 3) {
      last_updated = time(NULL);
      index = (index + 1) % 3;
   }

   if (mi->cinfo[ fields[index] ] != NULL)
      return mi->cinfo[ fields[index] ];
   else {
      /* draw the filename if field info not there, but trim it down */
      offset = 0;
      if (strlen(mi->filename) > 49)
         offset += strlen(mi->filename) - 49;

      return mi->filename + offset;
   }
}

/* given a number N, this builds and returns a printf(3) "%Ns" string */
char *
num2fmt(int n)
{
   static char format[255];

   if (n <= 0) {
      endwin();
      errx(1, "num2sfmt: invalid number %d provided", n);
   }

   snprintf(format, sizeof(format), "%%-%d.%ds", n, n);
   return format;
}

/* paint the status bar */
void
paint_status_bar()
{
   swindow    *w;
   static char scratchpad[500];
   char       *focusName;
   int         percent;

   /* XXX NOTE: to right-align the text we snprintf to the scratchpad above and
    * then use the cmd_row_fmt element of ui to right-align the string
    * (scratchpad) in the display.
    * this should be changed... just determine width of display...all fields
    * printed are fixed-width.
    */

   /* determine focus'd window name */
   if (ui.active == ui.library) {
      focusName = "library";
      w = ui.library;
   } else {
      focusName = "playlist";
      w = ui.playlist;
   }

   /* determine how far we've scrolled through the window */
   if (w->nrows == 0)
      percent = 100;
   else
      percent = 100 * (w->voffset + w->crow + 1) / w->nrows;

   /* build the string to print */
   snprintf(scratchpad, sizeof(scratchpad),
      "[%s%s%s] %6d,%-3d %3d%%",
      focusName,
      (ui.active == ui.library ? "" : ":"),
      (ui.active == ui.library ? "" : viewing_playlist->name),
      w->voffset + w->crow + 1,
      w->hoffset,
      percent);

   /* do the printing */
   werase(ui.command);
   wattron(ui.command, COLOR_PAIR(colors.status));
   mvwprintw(ui.command, 0, 0, ui.cmd_row_fmt, scratchpad);
   wattroff(ui.command, COLOR_PAIR(colors.status));
   wrefresh(ui.command);
}

/* paint the player */
void
paint_player()
{
   static char *playmode;
   static char *finfo;
   static int   in_hour;
   static int   in_minute;
   static int   in_second;
   static int   percent, whole;
   int w, h;

   getmaxyx(stdscr, h, w);

   /* if nothing's playing, a shameless plug */
   if (!mp_status.is_playing) {
      werase(ui.player);
      wattron(ui.player, COLOR_PAIR(colors.player));
      mvwprintw(ui.player, 0, 0, num2fmt(w), "vitunes...");
      wattroff(ui.player, COLOR_PAIR(colors.player));
      wrefresh(ui.player);
      return;
   }

   /* determine time into current selection */
   in_hour   = (int)  roundf(mp_status.position / 3600);
   in_minute = ((int) roundf(mp_status.position)) % 3600 / 60;
   in_second = ((int) roundf(mp_status.position)) % 60;

   /* determine percent time into current selection */
   percent = -1;
   if (playing_playlist->files[player.current]->plength_ms > 0) {
      whole = playing_playlist->files[player.current]->plength_ms / 1000000;
      percent = roundf(100.0 * mp_status.position / whole);
   }

   /* get character for playmode */
   switch (player.mode) {
      case MP_MODE_LINEAR:
         playmode = "-";
         break;
      case MP_MODE_LOOP:
         playmode = "O";
         break;
      case MP_MODE_RANDOM:
         playmode = "?";
         break;
   }

   /* determine info about song to show */
   finfo = player_get_field2show(playing_playlist->files[player.current]);

   /* draw */
   werase(ui.player);
   wattron(ui.player, COLOR_PAIR(colors.player));
   mvwprintw(ui.player, 0, 0, num2fmt(w), " "); /* this fill the bg color */
   mvwprintw(ui.player, 0, 0,
      "[%s] %8.8s +%2.2d:%2.2d:%2.2d (%d%%) %-49.49s",
      playmode,
      (player.paused ? "-PAUSED-" : ""),
      in_hour, in_minute, in_second,
      percent,
      finfo);
   wattroff(ui.player, COLOR_PAIR(colors.player));
   wrefresh(ui.player);
}

/* paint the library window */
void
paint_library()
{
   char *str;
   int   row, hoff, index;

   /* if library window is hidden, nothing to do */
   if (ui.library->cwin == NULL) return;

   werase(ui.library->cwin);
   wattron(ui.library->cwin, COLOR_PAIR(colors.library));

   for (row = 0; row < ui.library->h; row++) {

      index = ui.library->voffset + row;

      /* apply attributes */
      if (index < mdb.nplaylists && mdb.playlists[index] == playing_playlist)
         wattron(ui.library->cwin, COLOR_PAIR(colors.playing));

      if (index < mdb.nplaylists && mdb.playlists[index]->needs_saving)
         wattron(ui.library->cwin, A_BOLD);

      if (row == ui.library->crow)
         wattron(ui.library->cwin, A_REVERSE);

      if (index >= mdb.nplaylists)
         wattron(ui.library->cwin, COLOR_PAIR(colors.tildas));

      /* draw the row */
      if (index >= mdb.nplaylists)
         mvwprintw(ui.library->cwin, row, 0, "~");
      else {
         /* determine horizontal offset */
         str = mdb.playlists[index]->name;
         hoff = ui.library->hoffset;
         if (hoff >= strlen(str))
            hoff = strlen(str);

         /* draw it */
         mvwprintw(ui.library->cwin, row, 0,
            num2fmt(ui.library->w),
            str + hoff);
      }

      /* un-apply attributes */
      if (index >= mdb.nplaylists) {
         wattroff(ui.library->cwin, COLOR_PAIR(colors.tildas));
         wattron(ui.library->cwin, COLOR_PAIR(colors.library));
      }

      if (row == ui.library->crow) {
         wattroff(ui.library->cwin, A_REVERSE);
         wattron(ui.library->cwin, COLOR_PAIR(colors.library));
      }

      if (index < mdb.nplaylists && mdb.playlists[index]->needs_saving) {
         wattroff(ui.library->cwin, A_BOLD);
         wattron(ui.library->cwin, COLOR_PAIR(colors.library));
      }

      if (index < mdb.nplaylists && mdb.playlists[index] == playing_playlist) {
         wattroff(ui.library->cwin, COLOR_PAIR(colors.playing));
         wattron(ui.library->cwin, COLOR_PAIR(colors.library));
      }
   }

   wattroff(ui.library->cwin, COLOR_PAIR(colors.library));
   wrefresh(ui.library->cwin);
}

/* paint the playlist window */
void
paint_playlist()
{
   playlist   *plist;
   bool        hasinfo;
   char       *str;
   int         findex, row, col, colwidth;
   int         xoff, hoff, strhoff;
   int         cattr;


   plist = viewing_playlist;

   werase(ui.playlist->cwin);

   for (row = 0; row < ui.playlist->h; row++) {

      /* get index of file to show */
      findex = row + ui.playlist->voffset;

      /* apply row attributes */
       wattron(ui.playlist->cwin, COLOR_PAIR(colors.playlist));

      if (plist == playing_playlist && findex == player.current)
         wattron(ui.playlist->cwin, COLOR_PAIR(colors.playing));

      if (row == ui.playlist->crow)
         wattron(ui.playlist->cwin, A_REVERSE);

      if (findex >= plist->nfiles)
         wattron(ui.playlist->cwin, COLOR_PAIR(colors.tildas));

      /* draw the row */
      if (findex >= plist->nfiles)
         mvwprintw(ui.playlist->cwin, row, 0, "~");
      else {
         /* this acheives the A_REVERSE attribute spanning the entire row */
         mvwprintw(ui.playlist->cwin, row, 0, num2fmt(ui.playlist->w), " ");

         /* does the file have any meta-info? */
         hasinfo = false;
         for (col = 0; col < _mi_display.nfields; col++) {
            if (plist->files[findex]->cinfo[_mi_display.order[col]] != NULL)
               hasinfo = true;
         }

         /* if there's no meta info, just show filename */
         if (!hasinfo) {
            mvwprintw(ui.playlist->cwin, row, 0, num2fmt(ui.playlist->w),
               plist->files[findex]->filename);
         } else {

            /* loop through all fields of file and display each ... */
            xoff = 0;
            hoff = ui.playlist->hoffset;
            for (col = 0; col < _mi_display.nfields; col++) {

               /* is horizontal offset big enough to skip this field? */
               if (hoff >= _mi_display.widths[col]) {
                  hoff -= _mi_display.widths[col];
                  continue;
               }

               /* field shown off the screen? */
               if (xoff >= ui.playlist->w)
                  continue;

               /* get string to show (str) and horizontal offset to apply (shoff) */
               str = plist->files[findex]->cinfo[_mi_display.order[col]];
               if (str != NULL) {
                  if (hoff > strlen(str))
                     strhoff = strlen(str);
                  else
                     strhoff = hoff;
               }

               /* apply column attribute (only if file is NOT playing) */
               cattr = COLOR_PAIR(colors.cinfos[_mi_display.order[col]]);
               if ((plist != playing_playlist || findex != player.current)
               && colors.cinfos_set[_mi_display.order[col]])
                  wattron(ui.playlist->cwin, cattr);

               /* determine width of this field */
               colwidth = _mi_display.widths[col] - hoff;
               if (xoff + colwidth > ui.playlist->w)
                  colwidth = ui.playlist->w - xoff;

               /* print the column */
               mvwprintw(ui.playlist->cwin, row, xoff, num2fmt(colwidth),
                  (str == NULL ? " " : str + strhoff));

               /* un-apply column attribute */
               if ((plist != playing_playlist || findex != player.current)
               && colors.cinfos_set[_mi_display.order[col]]) {
                  wattroff(ui.playlist->cwin, cattr);
                  wattron(ui.playlist->cwin, COLOR_PAIR(colors.playlist));
               }

               xoff += 1 + colwidth; /* +1 for space between columns */
               hoff = 0;
            }
         }
      }

      /* un-apply row attributes */
      if (findex >= plist->nfiles)
         wattroff(ui.playlist->cwin, COLOR_PAIR(colors.tildas));

      if (row == ui.playlist->crow)
         wattroff(ui.playlist->cwin, A_REVERSE);

      if (plist == playing_playlist && findex == player.current)
         wattroff(ui.playlist->cwin, COLOR_PAIR(colors.playing));

      wattroff(ui.playlist->cwin, COLOR_PAIR(colors.playlist));
   }

   wrefresh(ui.playlist->cwin);
}

/* paint all windows */
void
paint_all()
{
   wattron(stdscr, COLOR_PAIR(colors.bars));
   ui_draw_borders();
   wattroff(stdscr, COLOR_PAIR(colors.bars));

   paint_player();
   paint_status_bar();
   paint_library();
   paint_playlist();
}

/* Paints an error message to the command/status window.  The usage is
 * identical to that of printf(3) (vwprintw(3) actually).
 */
void
paint_error(char *fmt, ...)
{
   va_list ap;

   werase(ui.command);
   wmove(ui.command, 0, 0);
   wattron(ui.command, COLOR_PAIR(colors.errors));

   va_start(ap, fmt);
   vwprintw(ui.command, fmt, ap);
   va_end(ap);

   beep();
   wattroff(ui.command, COLOR_PAIR(colors.errors));
   wrefresh(ui.command);
}

/* Paints an informational message to the command/status window.  The usage
 * is identical to that of printf(3) (vwprintw(3) actually).
 */
void
paint_message(char *fmt, ...)
{
   va_list ap;

   werase(ui.command);
   wmove(ui.command, 0, 0);
   wattron(ui.command, COLOR_PAIR(colors.messages));

   va_start(ap, fmt);
   vwprintw(ui.command, fmt, ap);
   va_end(ap);

   wattroff(ui.command, COLOR_PAIR(colors.messages));
   wrefresh(ui.command);
}

/* Each of these members of the global color object will be
 * run through init_pair(3)
 */
void
paint_setup_colors()
{
   int i;

   /* setup the indices to be used for init_pair/COLOR_PAIR */
   colors.bars = 1;
   colors.errors = 2;
   colors.messages = 3;
   colors.player = 4;
   colors.status = 5;
   colors.library = 6;
   colors.playlist = 7;
   colors.playing = 8;
   colors.tildas = 9;

   /* setup default colors */
   init_pair(colors.bars,     COLOR_WHITE, COLOR_BLACK);
   init_pair(colors.errors,   COLOR_WHITE, COLOR_RED);
   init_pair(colors.messages, COLOR_RED,   COLOR_BLACK);
   init_pair(colors.player,   COLOR_GREEN, COLOR_BLACK);
   init_pair(colors.status,   COLOR_WHITE, COLOR_BLACK);
   init_pair(colors.library,  COLOR_WHITE, COLOR_BLACK);
   init_pair(colors.playlist, COLOR_WHITE, COLOR_BLACK);
   init_pair(colors.playing,  COLOR_GREEN, COLOR_BLACK);
   init_pair(colors.tildas,   COLOR_BLUE,  COLOR_BLACK);

   /* colors for cinfo fields (columns in playlist window) */
   for (i = 0; i < MI_NUM_CINFO; i++) {
      colors.cinfos[i] = 10 + i;
      colors.cinfos_set[i] = false;
   }
}

int
paint_str2item(const char *str)
{
   int i;

   if (strcasecmp(str, "bars") == 0)
      return colors.bars;
   else if (strcasecmp(str, "errors") == 0)
      return colors.errors;
   else if (strcasecmp(str, "messages") == 0)
      return colors.messages;
   else if (strcasecmp(str, "player") == 0)
      return colors.player;
   else if (strcasecmp(str, "status") == 0)
      return colors.status;
   else if (strcasecmp(str, "library") == 0)
      return colors.library;
   else if (strcasecmp(str, "playlist") == 0)
      return colors.playlist;
   else if (strcasecmp(str, "playing") == 0)
      return colors.playing;
   else if (strcasecmp(str, "tildas") == 0)
      return colors.tildas;

   /* if reached here, check cinfo's array */
   for (i = 0; i < MI_NUM_CINFO; i++) {
      if (strcasecmp(str, MI_CINFO_NAMES[i]) == 0)
         return colors.cinfos[i];
   }

   return -1;
}

int
paint_str2color(const char *str)
{
   if (strcasecmp(str, "black") == 0)
      return COLOR_BLACK;
   else if (strcasecmp(str, "red") == 0)
      return COLOR_RED;
   else if (strcasecmp(str, "green") == 0)
      return COLOR_GREEN;
   else if (strcasecmp(str, "yellow") == 0)
      return COLOR_YELLOW;
   else if (strcasecmp(str, "blue") == 0)
      return COLOR_BLUE;
   else if (strcasecmp(str, "magenta") == 0)
      return COLOR_MAGENTA;
   else if (strcasecmp(str, "cyan") == 0)
      return COLOR_CYAN;
   else if (strcasecmp(str, "white") == 0)
      return COLOR_WHITE;
   else
      return -1;
}
